// Contrôle d'un relais bistable pour perforatrice de cartons
// PP le 04/02/2022

// Version M0 = le capteur de position de poinçon donne un 0 logique lorsqu'il est masqué

// La carte est alimentée en 5V par le connecteur PWM de la carte Grbl
// Ce 5 V est propagé sur la carte IBT-2 pour l'alim de son étage de commande

// Connecteurs du circuit imprimé :
// 4 broches vers la carte IBT2 : 5V, RPWM_Output, LPWM_Output, GND
// 3 broches vers le capteur de position du poinçon : 5V, pos_poincon, GND
// 5 broches vers la carte Grbl : 5V, G_Start, GND, cde, G_Pause
// Et vers le panneau interface : led_verte, led_rouge, B_init, B_cycle, GND

// A NOTER concernant les commandes par registres :
// -->> Tout ce qui peut réduire le temps de cycle contribue à réduire l'échauffement 
// La commande par les registres gagne environ 3.5 ms sur la durée de cycle 
// par rapport à des digitalRead/digitalWrite
// Le Serial.print prend environ 450 ns
// La durée réelle d'un cycle sans les Serial.print est d'environ 17200 ms sur la perfo #1
// Rappel :
// Port B : broches numériques 8 à 13
// Port C : broches d'entrée analogiques
// Port D : broches numériques 0 à 7
// Noms des registres : DDRp = Direction du port, PORTp = écriture, PINp = lecture

// 06/04/2023 : Nouvelle version avec une meilleure gestion des défauts pendant le cycle

// Les broches :
int cde = 2;          // Impulsion de commande (De la broche PWM sur la carte Grbl)
int G_Pause = 3;      // Vers la broche Pause (A1) de la carte Grbl
int pos_poincon = 4;  // Position du poinçon
int RPWM_Output = 5;  // Commande dans un sens
int LPWM_Output = 6;  // ... et dans l'autre
int G_Start = 7;      // Vers la broche Start (A2) de la carte Grbl
int B_cycle = 9;      // Bouton de déclenchement de cycle 
int B_init = 10;      // Bouton de ré-initialisation
int led_rouge = 11; 
int led_verte = 12;

boolean cycle_done, B_cycle_done, defaut, raz_def, init_done;
// Table de valeurs en cosinus
int table_cos[16] ={255,248,236,220,201,178,154,129,104,80,57,38,22,10,3,0};
//Durée allouée à un cycle (en micros) en sachant qu'à F6000 et accélération = 3000, on dispose de 45 ms
const unsigned long tcycle = 30000; 
 
void setup()
{
//  Serial.begin(115200); // Pour mesurer le temps de cycle
  // A ce stade il n'y a pas d'intérêt à configurer le sens des broches par les registres (gagner du temps ne sert à rien) 
  // Donc pour la lisibilité on reste classique :o)
  pinMode(RPWM_Output, OUTPUT);
  pinMode(LPWM_Output, OUTPUT);
  pinMode(G_Pause, OUTPUT);
  pinMode(G_Start, OUTPUT);
  pinMode(led_rouge, OUTPUT);
  pinMode(led_verte, OUTPUT);
  pinMode(cde, INPUT);
  pinMode(B_cycle, INPUT_PULLUP);
  pinMode(B_init, INPUT_PULLUP);
  pinMode(pos_poincon, INPUT);
  
  digitalWrite(LPWM_Output, 0);
  digitalWrite(RPWM_Output, 0);
  digitalWrite(G_Pause, HIGH);  
  digitalWrite(G_Start, HIGH);
  cycle_done = true;
  raz_def = false;
  init_done = false;
  defaut = false;
  init_cycle(); // Remonte le poinçon au cas où il serait en bas
}

// Exécution d'un cycle de perforation
void exec_cycle(){
  unsigned long time_in, diff, pas;
  time_in = micros();
  diff = 0;
  PORTB |= B00010000; // led_verte HIGH
  PORTD |= B00100000; // RPWM_Output HIGH
  while((PIND & B00010000) == B00000000 ){ // Attend que le carton soit transpercé
    diff = micros() - time_in;
    if (diff >= tcycle){ // Watchdog de descente
      halt_grbl();
      break; 
    }
  }
//  Serial.print("Temps de descente = "); Serial.println(diff);
  if (!defaut){
    PORTD &= B11011111; // RPWM_Output LOW
    PORTD |= B01000000; // LPWM_Output HIGH
    while((PIND & B00010000) == B00010000 ){  // Attend que l'on ressorte du carton
      diff = micros() - time_in;
      if (diff >= tcycle){ // Watchdog de sortie du carton
        halt_grbl();
        break; 
      }    
    }  
    // Décroissance sur 16 paliers variant en cosinus pour le même temps que : descente + sortie du carton
    if (!defaut){
      pas = diff/16;
      for (int i = 0; i<16; i++){
        analogWrite(LPWM_Output, table_cos[i]);
        delayMicroseconds(pas); 
      }
    }
  }
  // On coupe systématiquement toutes les commandes en fin de cycle pour couvrir les cas de défaut dont on sort par break
  PORTD &= B10011111; // LPWM et RPWM LOW
  PORTB &= B11101111; // led_verte LOW
  cycle_done = true;       
  B_cycle_done = true;
  diff = micros() - time_in; 
  if (((PIND & B00010000) == B00010000 ) || (diff >= tcycle)) halt_grbl(); // Watchdog sur la durée ou poinçon non remonté     
//  Serial.print("Temps total = "); Serial.println(diff);
}

// Stoppe l'exécution Grbl
void halt_grbl(){
  PORTB |= B00001000; // led_rouge HIGH
  PORTD &= B11110111; // Pause LOW
  defaut = true;
  // On tente de forcer une remontée de poinçon
  PORTD &= B11011111; // RPWM_Output LOW
  PORTD |= B01000000; // LPWM_Output HIGH
  delayMicroseconds(15000);
  PORTD &= B10011111; // LPWM et RPWM LOW
}

// Ré-initialisation
void init_cycle(){
  if (!raz_def){
    PORTD |= B01000000; // LPWM_Output HIGH
    delay(15);          // Impulsion de 15 ms pour remonter le poinçon 
    PORTD &= B10111111; // LPWM_Output LOW
    if ((PIND & B00010000) == B00000000 ){ // Si le poinçon est remonté alors on relance la carte Grbl    
      PORTB &= B11110111; // led_rouge LOW 
      PORTD |= B00001000; // Pause HIGH
      PORTD &= B01111111; // G_start LOW
      delay(10);          // Impulsion de 10 ms pour envoi du Start Grbl
      PORTD |= B10000000; // G_start HIGH
      defaut = false;     // Et on efface le défaut
      raz_def = true;
    }
    init_done = true;
  }
}
 
void loop(){
  // Commande par Grbl
  if ((PIND & B00000100) == B00000100){ // Passage à 1 de la commande
    if (!defaut){
      if (!cycle_done){
        exec_cycle();
      }
    }
  }else{
    cycle_done = false; // Au passage à 0 de la commande, on ré-arme pour le coup suivant 
  }
  
  // Commande par le bouton. 
  // Elle est distincte de la précédente pour introduire une tempo et ne pas rallonger 
  // le temps d'exécution de la commande Grbl pure par des tests, synonyme d'échauffement supplémentaire
  if ((PINB & B00000010) == B00000000){ // Appui sur le bouton
    if (((PIND & B00010000) == B00000000)&&(!defaut)){ // Seulement si poincon haut et pas défaut
      if (!B_cycle_done){
        exec_cycle();
        delay(200);
      }           
    }else{
      halt_grbl();
    }
  }else{
    B_cycle_done = false; // A la remontée du bouton, on ré-arme pour le coup suivant 
  }

  // Bouton de ré-initialisation
  if ((PINB & B00000100) == B00000000){
    if (!init_done){
      raz_def = false;
      init_cycle();
    }
  }else{  // Remontée du bouton pour ne faire qu'une init à la fois
    init_done = false;
  }
  // Pour régler la position du capteur
//  if (digitalRead(pos_poincon) == LOW) digitalWrite(led_rouge, LOW); else digitalWrite(led_rouge, HIGH);
}
